# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mars/serialize/protos/operand.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from mars.serialize.protos import value_pb2 as mars_dot_serialize_dot_protos_dot_value__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='mars/serialize/protos/operand.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n#mars/serialize/protos/operand.proto\x1a!mars/serialize/protos/value.proto\"\x80\x01\n\nOperandDef\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12#\n\x04\x61ttr\x18\x03 \x03(\x0b\x32\x15.OperandDef.AttrEntry\x1a\x33\n\tAttrEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Value:\x02\x38\x01\x62\x06proto3'
  ,
  dependencies=[mars_dot_serialize_dot_protos_dot_value__pb2.DESCRIPTOR,])




_OPERANDDEF_ATTRENTRY = _descriptor.Descriptor(
  name='AttrEntry',
  full_name='OperandDef.AttrEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='OperandDef.AttrEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='OperandDef.AttrEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=152,
  serialized_end=203,
)

_OPERANDDEF = _descriptor.Descriptor(
  name='OperandDef',
  full_name='OperandDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='OperandDef.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='OperandDef.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='attr', full_name='OperandDef.attr', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_OPERANDDEF_ATTRENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=75,
  serialized_end=203,
)

_OPERANDDEF_ATTRENTRY.fields_by_name['value'].message_type = mars_dot_serialize_dot_protos_dot_value__pb2._VALUE
_OPERANDDEF_ATTRENTRY.containing_type = _OPERANDDEF
_OPERANDDEF.fields_by_name['attr'].message_type = _OPERANDDEF_ATTRENTRY
DESCRIPTOR.message_types_by_name['OperandDef'] = _OPERANDDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

OperandDef = _reflection.GeneratedProtocolMessageType('OperandDef', (_message.Message,), {

  'AttrEntry' : _reflection.GeneratedProtocolMessageType('AttrEntry', (_message.Message,), {
    'DESCRIPTOR' : _OPERANDDEF_ATTRENTRY,
    '__module__' : 'mars.serialize.protos.operand_pb2'
    # @@protoc_insertion_point(class_scope:OperandDef.AttrEntry)
    })
  ,
  'DESCRIPTOR' : _OPERANDDEF,
  '__module__' : 'mars.serialize.protos.operand_pb2'
  # @@protoc_insertion_point(class_scope:OperandDef)
  })
_sym_db.RegisterMessage(OperandDef)
_sym_db.RegisterMessage(OperandDef.AttrEntry)


_OPERANDDEF_ATTRENTRY._options = None
# @@protoc_insertion_point(module_scope)
